
/*----------------------------------------------------------------------------------------
|
| 	[ATMega88 mit SI570, LCD und Drehimpulsgeber]
| 	Copyright (C) [2010]  [Clemens Verstappen DL3ETW] 
| 
| 
| 	Basierendes Werk "TR2200GX with Silabs Si570 LO control firmware" von 
| 	Wulf-Gerd Traving, DL1FAC 2008, unter Verwendung der lcd.c und twimaster.c 
|	Module von Peter Fleury.
|
|
|	Funktionsumfang:
|	
|	Die im Flashmemory abgelegten Frequenzen (32Bit Integer) werden via I2C 
|	an den SI570 ausgegeben und ber das LCD Display mit Frequenz und Kanal 
|	angezeigt. Die Frequenzauswahl erfolgt ber den Drehimpulsgeber. 
|	Als Beispiel ist eine Frequenztabelle des Seefunks implementiert.
|
|	Fusebits:
|	Ext. Crystal Osc. 8.0-    MHz; Start-up time PWRDWN/RESET: 16K CK/14 CK + 65 ms
|	bei Verwendung eines 20MHz Quarzes.
|
|-----------------------------------------------------------------------------------------
| License:
| This program is free software; you can redistribute it and/or modify it under
| the terms of the GNU General Public License as published by the Free Software
| Foundation; either version 2 of the License, or (at your option) any later
| version.
| This program is distributed in the hope that it will be useful, but
|
| WITHOUT ANY WARRANTY;
|
| without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
| PURPOSE. See the GNU General Public License for more details.
|
| You should have received a copy of the GNU General Public License along with
| this program; if not, write to the Free Software Foundation, Inc., 51
| Franklin St, Fifth Floor, Boston, MA 02110, USA
|
| http://www.gnu.de/gpl-ger.html
`-----------------------------------------------------------------------------------------*/


#include "i2cmaster.h"			//	I2C Ansteuerung des SI570 (in Hardware)  				
#include "lcd.h"				//	Ansteuerung des zweireihigen LCD Displays
#include <avr/interrupt.h>		//	Fr Timer und den externen Interrupt 
#include <avr/pgmspace.h>		//	Lesen aus dem Flash, Progammspeicher
#include <avr/stdlib.h>			// 	Routinen zur Zeichenausgabe ber das LCD
#include <string.h>				//	Routine zur Ermittlung der Zeichenlnge
#include <util/delay.h>			//	Routine fr Zeitverzgerung

#define SI570ADDR 0xaa			//	I2C Adresse des SI570
#define FOUT0 56.3199			//	Hier die Frequenz (MHz) eingeben, die direkt nach Anlegen
								//	der Versorgungsspannung am SI570 erzeugt wird

const float FDCO_MAX = 5670.0; 					// MHz,  Si57x's fdco range -- DO NOT EDIT
const float FDCO_MIN = 4850.0; 					// MHz
const unsigned long int FDCO_MAXL = 5537109; 	// * 1024 Hz
const unsigned long int FDCO_MINL = 4736328; 	// * 1024 Hz
const unsigned char HS_DIV[6] = {11, 9, 7, 6, 5, 4};

uint8_t sreg,initial_n1,initial_hsdiv,kanal,zeiger=9, s[16];
uint8_t zeile1[16], zeile2[16],reg[6];

uint32_t initial_rfreq_long,frequenz;
uint32_t PROGMEM channel_tab [] PROGMEM = {
			1,	156050000, 160650000,			// Kanal, Schiffsfunkstelle, Ortsfeste Funkstelle
			2,	156100000, 160700000,
			3,	156150000, 160750000,
			4,	156200000, 160800000,
			5,	156250000, 160850000,
			6,	156300000, 156300000,
			7,	156350000, 160950000,
			8,	156400000, 156400000,
			9,	156450000, 156450000,
			10,	156500000, 156500000,
			11,	156550000, 156550000,
			12, 156600000, 156600000,
			13,	156650000, 156650000,
			14,	156700000, 156700000,
			15,	156750000, 156750000,
			16,	156800000, 156800000,
			17,	156850000, 156850000,
			18, 156900000, 161500000,
			19, 156950000, 161550000,
			20,	157000000, 161600000,
			21,	157050000, 161650000,
			22,	157100000, 161700000,
			23,	157150000, 161750000,
			24,	157200000, 161800000,
			25,	157250000, 161850000,
			26,	157300000, 161900000,
			27,	157350000, 161950000,
			28,	157400000, 162000000,
			60,	156025000, 160625000,
			61,	156075000, 160675000,
			62,	156125000, 160725000,
			63,	156175000, 160775000,
			64,	156225000, 160825000,
			65,	156275000, 160875000,
			66,	156325000, 160925000,
			67,	156375000, 156375000,
			68,	156425000, 156425000,
			69,	156475000, 156475000,
			70,	156525000, 156525000,
			71,	156575000, 156575000,
			72,	156625000, 156625000,
			73,	156675000, 156675000,
			74,	156725000, 156725000,
			75,	156775000, 156775000,
			76,	156825000, 156825000,
			77,	156875000, 156875000,
			78,	156925000, 161525000,
			79,	156975000, 161575000,
			80,	157025000, 161625000,
			81,	157075000, 161675000,
			82,	157125000, 161725000,
			83,	157175000, 161775000,
			84,	157225000, 161825000,
			85,	157275000, 161875000,
			86,	157325000, 161925000,
			87,	157375000, 157375000,
			88,	157425000, 157425000,
			89,	161975000, 161975000,		// AIS
			90,	162025000, 162025000,		// AIS
};

volatile struct BitVariablen{				//	um Status zu merken bei nderungen
    int kanal_update 	:1;
	int	freq_update  	:1;
	int	silabs_detected :1;
}status = {1};


SIGNAL(TIMER1_COMPA_vect) {					// Signal(Timer_Compa_vect) Interrupt jede 1/4 Sekunde
	PORTB ^= (1 << PORTB1);					// Lebenslicht 0.5 Hz
	status.freq_update = 1;
} 
	
SIGNAL(INT1_vect) {							// externer  Interrupt, steigende Flanke von Drehencoder B 
		status.kanal_update = 1;			// nderung in dieser Bitvariable merken um LCD+VCO updaten
		if (PINC & (1 << PC3)) 	{			// Wie ist Pegel von Drehencoder A?
			if (zeiger > 0)					// wenn noch grer eins, dann darf noch decrementiert werden
					zeiger--;				// Drehrichtung gegen den Uhrzeigersinn
			else
			zeiger=58;						// und oben wieder anfangen
			}
		else					{
			if (zeiger < 58)				// Der Pegel von Drehencoder A ist negativ
					zeiger++;				// Drehrichtung mit den Uhrzeigersinn
			else
			zeiger=0;						// und unten wieder anfangen
			}
		sreg = SREG;						// Timer1 aus Null setzen, um Drehen an Impulsgeber zu ermglichen  
		cli();								// Disable interrupts
		TCNT1 = 0x0000;						// Clear Timer1 auf, von null wieder anfangen 
		SREG = sreg;						// Restore global interrupt flag
		sei();
}

int I2C_ByteWrite (unsigned char reg, unsigned char data) {
	if (i2c_start (SI570ADDR|I2C_WRITE)) {	// failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{								// issuing start condition ok, device accessible */
		i2c_write(reg);
		i2c_write(data);
		i2c_stop();
	}
	return 0;
}

int I2C_regsWrite (void) {
    unsigned char i;
	if (i2c_start (SI570ADDR|I2C_WRITE)) 	{ // failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{								// issuing start condition ok, device accessible */
		i2c_write (7);						// load the new values into the device at registers 7 to 12
	    for (i=0; i<sizeof(reg); i++) i2c_write (reg[i]);
		i2c_stop();
	}
	return 0;
}

unsigned char I2C_ByteRead (unsigned char reg) {
	unsigned char ret;
	if (i2c_start (SI570ADDR|I2C_WRITE)) {	 // failed to issue start condition, possibly no device found */
		i2c_stop();
		status.silabs_detected = 0;
		return -1;
	}
	else 	{
		i2c_write (reg);
		i2c_rep_start (SI570ADDR+I2C_READ);  // set device address and read mode
		ret = i2c_readNak ();                // read one byte
		i2c_stop ();
		return ret;
	}
}

void ReadStartUpConfiguration(void)	{		// Reads start-up register contents for RFREQ, HS_DIV, and N1 and calculates the internal crystal frequency (xtal)
	unsigned char counter;					// recall the initial state of the device from nvm into ram
	if (I2C_ByteWrite(135, 0x01) < 0) 	{
		status.silabs_detected = 0;
		return;
	}										// read registers 7 to 12, reg[0] is equivalent to register 7 in the device, reg[5] is register 12 in the device
	for(counter=0; counter<6; counter++) 	{
		reg[counter] = I2C_ByteRead(counter+7);
	}
	initial_hsdiv = ((reg[0] & 0xE0) >> 5) + 4; // HS_DIV conversion, get reg 7 bits 5, 6, 7
												// initial_hsdiv's value could be verified here to ensure that it is one
												// of the valid HS_DIV values from the datasheet.
												// initial_n1 conversion
	initial_n1 = (( reg[0] & 0x1F ) << 2 ) + (( reg[1] & 0xC0 ) >> 6 );  // get reg 7 bits 0 to 4
												// add with reg 8 bits 7 and 8
	if(initial_n1 == 0) 	{
		initial_n1 = 1;
	}
	else if ((initial_n1 & 1) != 0) 	{		// add one to an odd number
		initial_n1 = initial_n1 + 1;			// RFREQ conversion using unsigned long
	}
	initial_rfreq_long = ( reg[1] & 0x3F );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[2] );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[3] );
	initial_rfreq_long = (initial_rfreq_long << 8) + ( reg[4] );
	initial_rfreq_long = (initial_rfreq_long << 6) + ( reg[5] >> 2 ); //ignore lowest two bits to fit into long
}

void silabs_setNewFreq (uint32_t newFrequency) {	// Program a new frequency
	unsigned char counter;
	unsigned char n1;
	unsigned char hsdiv;
	unsigned char validCombo;
	unsigned char reg137, reg135;
	unsigned int divider_max;
	unsigned int curr_div;
	unsigned int rn1;
	float currentFrequency;
	float ratio = 0;
	unsigned long final_rfreq_long;

	if (status.silabs_detected) 	{
		currentFrequency = (float) newFrequency / 1000000.0;	// find dividers (get the max and min divider range for the HS_DIV and N1 combo)
		divider_max = FDCO_MAXL / (newFrequency >> 10);
		curr_div = (FDCO_MINL / (newFrequency >> 10)) + 1;
		validCombo = 0;
		n1 = initial_n1;
		hsdiv = initial_hsdiv;
		while (curr_div <= divider_max) {				// check all the HS_DIV values with the next curr_div
			for(counter=0; counter<6; counter++) {		// get the next possible n1 value
				hsdiv = HS_DIV[counter];
				rn1 = curr_div % hsdiv;
				if (rn1 == 0) {							// no remainder, this is a possible n1 value
					n1 = curr_div / hsdiv;
					if ((n1 == 1) || ((n1 & 1) == 0)) goto foundCombo;
				}										// increment curr_div to find the next divider
		}												// since the current one was not valid
		curr_div += 1;
		}
		foundCombo:
				// if(validCombo == 0) at this point then there's an error
				// in the calculatio. Check if the provided fout0 and fout1
				// are not valid frequencies
				// calculate RFREQ organizing the float variables to save precision;
				// RFREQ is kept as an unsigned long
				// only 32 bits are available in the long format
				// RFREQ in the device has 34 bits of precision
				// only 34 of the 38 bits are needed since RFREQ is between 42.0 and 50.0 for xtal of 114.285MHz (nominal)
		ratio = currentFrequency / FOUT0;
		ratio *= ((float) n1 / (float) initial_n1) * ((float) hsdiv / (float) initial_hsdiv);
		final_rfreq_long = ratio * (float) initial_rfreq_long;
		reg[0] = 0;										// set the top 3 bits of reg 13
		reg[0] = (hsdiv-4) << 5;						// convert new N1 to the binary representation
		if (n1 == 1) n1 = 0;
		else if((n1 & 1) == 0) n1 = n1 - 1; 			//if n1 is even, subtract one
		reg[0] = (reg[0] & 0xE0) | (n1 >> 2);			// set reg 7 bits 0 to 4
		reg[1] = (n1 & 3) << 6;							// set reg 8 bits 6 and 7

		reg[1] |= final_rfreq_long >> 30;				// load new RFREQ into register map
		reg[2] = final_rfreq_long >> 22;
		reg[3] = final_rfreq_long >> 14;
		reg[4] = final_rfreq_long >> 6;
		reg[5] = final_rfreq_long << 2;
		
		do {								// Load the new frequency, first wait for previous setup beeing through
			reg135 = I2C_ByteRead(135);
		} while (reg135 & 0x40);			// get the current state of register 137
		reg137 = I2C_ByteRead(137);			// set the Freeze DCO bit in that register
		I2C_ByteWrite(137, reg137 | 0x10); 	// load the new values into the device at registers 7 to 12
		I2C_regsWrite ();					// get the current state of register 137
		reg137 = I2C_ByteRead(137);			// clear the FZ_DCO bit in that register
		I2C_ByteWrite(137, reg137 & 0xEF);	// set the NewFreq bit, bit will clear itself once the device is ready
		I2C_ByteWrite(135, 0x40);
	}
}

int main(void) 	{ 
	DDRB 	= 0b00000010;					// PB1 ist Lebenslicht, sonst Eingnge
	DDRC 	= 0b00110000;					// PortC ist Input, bis auf SCL, SDA	
	
	OCR1A	= 0x2626;						// Compare Match Register Laden, 4C4B = 1s ;2626 1313 = 250ms
	TCCR1B	= 0b00001101;					// WGM12 setzen, clear on compare match, divide by 1024
	TIMSK1	= 0b00000010;					// Timer1 compare match 1 interrupt enable
  	
	EICRA	|= (1<<ISC11)|(1<<ISC10);		// Steigende Flanke fr externen Interrupt 1, Drehimpulsgeber B
	EIMSK	|= (1<<INT1);					// externen Interrupt1 freigeben	

 	i2c_init();                             // initialize I2C library
	lcd_init(LCD_DISP_ON);					// LCD initialiesieren
 	lcd_clrscr();							// und lschen
    status.silabs_detected = 1;
    ReadStartUpConfiguration();
	_delay_ms (500.0);						// 	hier mal warten
	sei ();									// 	nun Interrupts freigeben

	for (;;)  	{
		if (status.kanal_update) 	{		// 	nur abarbeiten, wenn neuer Kanal angewhlt
			unsigned char index=0,z,i=0;
			kanal = pgm_read_dword (&channel_tab [zeiger*3]);		// den Kanal aus PGM lesen
			frequenz = pgm_read_dword (&channel_tab [zeiger*3+1]); 	// nun zugehrige Frequenz einlesen
			itoa(kanal, s, 10);				// 	hier kanal integer to ascii in s ablegen
			z=strlen(s);					//	wie lang ist s geworden?
				if (z==1) 	{				//	nach rechts schieben, wenn eine Ziffer
   				s[i+1]=s[i];
				s[i]=' ';
				}
		zeile1[index]=s[0];					// 	kopiere in Zeile1
  		index++;
		zeile1[index]=s[1];
  		index++;
		zeile1[index]=0;

		lcd_gotoxy(0,0);					// 	gehe auf Zeile1
  		lcd_puts("Kanal: ");
		for(i = 0; zeile1[i] && i < 8; i++)	// 	auslesen bis NULL, 8 Zeichen nicht mehr
  		lcd_putc(zeile1[i]);
											// 	Zweite Zeile
		ultoa(frequenz,s,10);				// 	32Bit unsign Integer in Ascii 
		for(i = 0; i < 10; i++)				// 	maximal 10Stellen kopieren in Zeile2
		zeile2[i]=s[i];
		lcd_gotoxy(0,1);					// 	dies in die zweite Zeile des LCD
		lcd_puts("Freq: ");	
		for(i = 0; zeile2[i] && i < 10; i++)// 	auslesen bis NULL, 8 Zeichen nicht mehr
  		lcd_putc(zeile2[i]);
		
		silabs_setNewFreq (frequenz-10700000);	// nun die ausgewhlte Frequenz zum SI570
												// hier wurden 10,7 MHz ZF abgezogen.
		status.kanal_update = 0;
		status.freq_update = 0;
		}
	}
}

	// Belegung der Ports
	// Port B Bit 0 PB0 (PCINT0/CKL0/ICP1)											
	//        Bit 1 PB1 (OC1A/PCINT1)			Lebenslicht
	//        Bit 2 PB2	(SS/OC1B/PCINT3)		-
  	//        Bit 3 PB3	(MOSI/OC2A/PCINT3))		ISP MOSI
  	//        Bit 4 PB4	(MISO/PCINT4))			ISP MISO
  	//        Bit 5 PB5 (SCK/PCINT5)			ISP SCK
  	//        Bit 6 PB6	(PCINT6/XTAL1/TOSC1)	20MHz X-TAL
	//		  Bit 7 PB7	(PCINT7/XTAL2/TOSC2)	20MHz X-TAL

 	// Port C Bit 0 PC0 (ADC0/PCINT8)			-
  	//        Bit 1 PC1	(ADC1/PCINT9)			-
 	//        Bit 2 PC2	(ADC2/PCINT10)			Druckknopf Drehimpulsgeber, not used
  	//        Bit 3 PC3	(ADC3/PCINT11)			Drehimpulsgeber A
  	//        Bit 4 PC4	(ADC4/SDA/PCINT12)		I2C SDA 
  	//        Bit 5 PC5	(ADC5/SCL/PCINT13)		I2C CLK 
  	//        Bit 6 PC6	(PCINT/RESET)			Reset
  			          										
	// Port D Bit 0 PD0	(PCINT16/RXD)			LCD E			
	//        Bit 1 PD1	(PCINT17/TXD)			LCD R/W
  	//        Bit 2 PD2	(PCINT18/INT0)			LCD RS
  	//        Bit 3 PD3	(PCINT19/OC2B/INT1)		Drehimpulsgeber B
  	//        Bit 4 PD4	(PCINT20/XCK/T0)		LCD Data dB4 	
  	//        Bit 5 PD5	(PCINT21/OC0B/T1)		LCD Data dB5 	
  	//        Bit 6 PD6	(PCINT22/OC0A/AIN0)		LCD Data dB6 	
  	//        Bit 7 PD7	(PCINT23/AIN1)			LCD Data dB7 			

/*
	reg[0] = 0b00000001; // 156,5MHz  0x01, Hs=4 N1=8 fr 156,5 MHz mit RFREQ 0x2 bdac 5aa4
	reg[1] = 0b11000010;	// alt 0xc2	
	reg[2] = 0xbd; 	//new_reference_rfeq >> 24;	
	reg[3] = 0xac; 	//new_reference_rfeq >> 16;	
	reg[4] = 0x5a;	//new_reference_rfeq >> 8;	
	reg[5] = 0xa4;	//new_reference_rfeq;
*/
